# -*- coding: utf-8 -*-

import logging

from .exceptions import PoyoException
from .parser import parse_string

__author__ = 'Raphael Pierzina'
__email__ = 'raphael@hackebrot.de'
__version__ = '0.4.1'

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = ['parse_string', 'PoyoException']



# -*- coding: utf-8 -*-

import functools
import logging

from ._nodes import Root, Section, Simple

from .exceptions import (
    NoMatchException, NoParentException, NoTypeException, IgnoredMatchException
)

from .patterns import (
    COMMENT, BLANK_LINE, DASHES, LIST, SIMPLE, SECTION,
    LIST_ITEM, NULL, TRUE, FALSE, FLOAT, INT, STR,
)

logger = logging.getLogger(__name__)


def log_callback(wrapped_function):
    """Decorator that produces DEBUG level log messages before and after
    calling a parser method.

    If a callback raises an IgnoredMatchException the log will show 'IGNORED'
    instead to indicate that the parser will not create any objects from
    the matched string.

    Example:
        DEBUG:poyo.parser:parse_simple <-     123: 456.789
        DEBUG:poyo.parser:parse_int <- 123
        DEBUG:poyo.parser:parse_int -> 123
        DEBUG:poyo.parser:parse_float <- 456.789
        DEBUG:poyo.parser:parse_float -> 456.789
        DEBUG:poyo.parser:parse_simple -> <Simple name: 123, value: 456.789>
    """

    def debug_log(message):
        """Helper to log an escaped version of the given message to DEBUG"""
        logger.debug(message.encode('unicode_escape').decode())

    @functools.wraps(wrapped_function)
    def _wrapper(parser, match, **kwargs):
        func_name = wrapped_function.__name__

        debug_log(u'{func_name} <- {matched_string}'.format(
            func_name=func_name,
            matched_string=match.group(),
        ))

        try:
            result = wrapped_function(parser, match, **kwargs)
        except IgnoredMatchException:
            debug_log(u'{func_name} -> IGNORED'.format(func_name=func_name))
            raise

        debug_log(u'{func_name} -> {result}'.format(
            func_name=func_name,
            result=result,
        ))

        return result
    return _wrapper


class _Parser(object):
    def __init__(self, source):
        self.pos = 0
        self.source = source
        self.max_pos = len(self.source)

        self.root = Root()
        self.seen = [self.root]

        self.rules = (
            (COMMENT, self.parse_comment),
            (BLANK_LINE, self.parse_blankline),
            (DASHES, self.parse_dashes),
            (LIST, self.parse_list),
            (SIMPLE, self.parse_simple),
            (SECTION, self.parse_section),
        )

        self.tag_rules = (
            (NULL, self.parse_null),
            (TRUE, self.parse_true),
            (FALSE, self.parse_false),
            (INT, self.parse_int),
            (FLOAT, self.parse_float),
            (STR, self.parse_str),
        )

    def find_at_level(self, level):
        for candidate in reversed(self.seen):
            if candidate.level < level:
                return candidate
        raise NoParentException(
            'Unable to find element at level {}'.format(level)
        )

    def read_from_tag(self, string):
        for pattern, callback in self.tag_rules:
            match = pattern.match(string)

            if not match:
                # Pattern does not match, try the next one
                continue

            if match.group() != string:
                # Unable to match the full string, try the next pattern
                continue

            return callback(match)

        error_message = 'Unable to determine type for "{}"'
        raise NoTypeException(error_message.format(string))

    @log_callback
    def parse_comment(self, match):
        """Ignore line comments."""
        raise IgnoredMatchException

    @log_callback
    def parse_blankline(self, match):
        """Ignore blank lines."""
        raise IgnoredMatchException

    @log_callback
    def parse_dashes(self, match):
        """Ignore lines that contain three dash symbols."""
        raise IgnoredMatchException

    @log_callback
    def parse_list(self, match):
        groups = match.groupdict()
        level = len(groups['indent'])
        parent = self.find_at_level(level)

        item_matches = LIST_ITEM.findall(groups['items'])

        list_items = [
            self.read_from_tag(value)
            for value in item_matches
        ]

        variable = self.read_from_tag(groups['variable'])
        return Simple(variable, level, list_items, parent=parent)

    @log_callback
    def parse_simple(self, match):
        groups = match.groupdict()

        level = len(groups['indent'])
        parent = self.find_at_level(level)

        variable = self.read_from_tag(groups['variable'])
        value = self.read_from_tag(groups['value'])

        return Simple(variable, level, value, parent=parent)

    @log_callback
    def parse_section(self, match):
        groups = match.groupdict()

        level = len(groups['indent'])
        parent = self.find_at_level(level)

        return Section(
            self.read_from_tag(groups['variable']),
            level,
            parent=parent
        )

    @log_callback
    def parse_null(self, match):
        return None

    @log_callback
    def parse_true(self, match):
        return True

    @log_callback
    def parse_false(self, match):
        return False

    @log_callback
    def parse_int(self, match):
        return int(match.group())

    @log_callback
    def parse_float(self, match):
        return float(match.group())

    @log_callback
    def parse_str(self, match):
        quotes = match.group('quotes')
        return match.group().strip(quotes)

    def find_match(self):
        """Try to find a pattern that matches the source and calll a parser
        method to create Python objects.

        A callback that raises an IgnoredMatchException indicates that the
        given string data is ignored by the parser and no objects are created.

        If none of the pattern match a NoMatchException is raised.
        """
        for pattern, callback in self.rules:
            match = pattern.match(self.source, pos=self.pos)

            if not match:
                continue

            try:
                node = callback(match)
            except IgnoredMatchException:
                pass
            else:
                self.seen.append(node)

            return match

        raise NoMatchException(
            'None of the known patterns match for {}'
            ''.format(self.source[self.pos:])
        )

    def __call__(self):
        """Parse the given string data and sequentually update the current
        cursor position until the end is reached.

        Return the Root object if successful.
        """
        while self.pos < self.max_pos:
            match = self.find_match()
            self.pos = match.end()
        return self.root()


def parse_string(string):
    parser = _Parser(string)
    return parser()



# -*- coding: utf-8 -*-

import codecs
import pytest

from poyo import parse_string


@pytest.fixture
def string_data(ymlfile):
    filename = 'tests/{}.yml'.format(ymlfile)
    with codecs.open(filename, encoding='utf-8') as fh:
        return fh.read()


@pytest.mark.parametrize('ymlfile', ['foobar'])
def test_parse_string(string_data):
    expected = {
        u'default_context': {
            u'greeting': u'こんにちは',
            u'email': u'raphael@hackebrot.de',
            u'docs': True,
            u'gui': False,
            u'lektor': '0.0.0.0:5000',
            u'relative-root': '/',
            123: 456.789,
            u'some:int': 1000000,
            u'foo': u'hallo #welt',
            u'trueish': u'Falseeeeeee',
            u'doc_tools': [u'mkdocs', u'sphinx', None],
        },
        u'zZz': True,
        u'NullValue': None,
        u'Hello World': {
            None: u'This is madness',
            u'gh': u'https://github.com/{0}.git',
        },
        u'Yay #python': u'Cool!'
    }

    assert parse_string(string_data) == expected


@pytest.mark.parametrize('ymlfile', ['no-newline'])
def test_parse_string_no_newline(string_data):
    expected = {
        u'Hello World': {
            u'name': u'Toni Chu',
            u'gh': u'https://github.com/{0}.git',
        },
        u'Yay #python': 3.6,
    }
    assert parse_string(string_data) == expected


@pytest.mark.parametrize('ymlfile', ['no-newline-list'])
def test_parse_string_no_newline_list(string_data):
    expected = {
        u'Hello World': {
            u'numbers': [
                1,
                2,
            ],
            u'name': u'Toni Chu',
            u'gh': u'https://github.com/{0}.git',
            u'doc_tools': [
                u'mkdocs',
                u'sphinx',
                None,
            ]
        },
    }
    assert parse_string(string_data) == expected



# -*- coding: utf-8 -*-

import pytest

from poyo import patterns


def test_COMMENT_matches_a_comment_line():
    match = patterns.COMMENT.match('# baz\n - bar\nfoobar')
    assert match is not None
    assert match.group() == '# baz\n'


def test_COMMENT_matches_a_comment_line_with_leading_whitespace():
    match = patterns.COMMENT.match('    # baz\n - bar\nfoobar')
    assert match is not None
    assert match.group() == '    # baz\n'


def test_BLANK_LINE_matches_a_blank_line():
    match = patterns.BLANK_LINE.match('  \nfoobar')
    assert match is not None
    assert match.group() == '  \n'


def test_DASHES_matches_document_separator():
    match = patterns.DASHES.match('---\nfoobar')
    assert match is not None
    assert match.group() == '---\n'


def test_LIST_matches_a_list():
    match = patterns.LIST.match('hmm:\n  - foo\n - "bar"\n    - 123\n-bing')
    assert match is not None
    assert match.group() == 'hmm:\n  - foo\n - "bar"\n    - 123\n'

    groups = match.groupdict()
    assert groups['indent'] == ''
    assert groups['items'] == '  - foo\n - "bar"\n    - 123\n'
    assert groups['variable'] == 'hmm'


def test_LIST_matches_a_list_with_blank_lines():
    match = patterns.LIST.match(
        'hmm:\n  - foo\n\n    - "bar"\n    - 123\n-bing'
    )
    assert match is not None
    assert match.group() == 'hmm:\n  - foo\n\n    - "bar"\n    - 123\n'

    groups = match.groupdict()
    assert groups['indent'] == ''
    assert groups['items'] == '  - foo\n\n    - "bar"\n    - 123\n'
    assert groups['variable'] == 'hmm'


def test_LISTITEM_finds_all_listitems():
    found = patterns.LIST_ITEM.findall(' - foo # baz\n - bar\nfoobar')
    assert found is not None
    assert found == ['foo', 'bar']


def test_LISTITEM_finds_all_listitems_with_blank_lines():
    found = patterns.LIST_ITEM.findall(' - foo # baz\n\n - bar\nfoobar')
    assert found is not None
    assert found == ['foo', 'bar']


def test_SIMPLE_matches_a_simple_key_value_line():
    match = patterns.SIMPLE.match('   foo: bar # baz\n - bar\nfoobar')
    assert match is not None
    assert match.group() == '   foo: bar # baz\n'

    groups = match.groupdict()
    assert groups['indent'] == '   '
    assert groups['variable'] == 'foo'
    assert groups['value'] == 'bar'


def test_SECTION_matches_a_section_key_line():
    match = patterns.SECTION.match('   foo: # baz\n      - bar\nfoobar')
    assert match is not None
    assert match.group() == '   foo: # baz\n'

    groups = match.groupdict()
    assert groups['indent'] == '   '
    assert groups['variable'] == 'foo'


@pytest.mark.parametrize('null_value', ['null', 'Null', 'NULL', '~'])
def test_NULL_matches_a_null_value(null_value):
    match = patterns.NULL.match(null_value)
    assert match is not None
    assert match.group() == null_value


@pytest.mark.parametrize('true_value', ['true', 'True', 'TRUE'])
def test_TRUE_matches_a_true_value(true_value):
    match = patterns.TRUE.match(true_value)
    assert match is not None
    assert match.group() == true_value


@pytest.mark.parametrize('false_value', ['false', 'False', 'FALSE'])
def test_FALSE_matches_a_false_value(false_value):
    match = patterns.FALSE.match(false_value)
    assert match is not None
    assert match.group() == false_value


@pytest.mark.parametrize('float_value', ['1.0', '42.820308', '39.2234e9'])
def test_FLOAT_matches_a_float_value(float_value):
    match = patterns.FLOAT.match(float_value)
    assert match is not None
    assert match.group() == float_value


@pytest.mark.parametrize('int_value', ['1', '42', '3922349'])
def test_INT_matches_an_int_value(int_value):
    match = patterns.INT.match(int_value)
    assert match is not None
    assert match.group() == int_value


def test_STR_matches_a_string_value():
    match = patterns.STR.match('foo bar baz ')
    assert match is not None
    assert match.group() == 'foo bar baz '

    groups = match.groupdict()
    assert groups['quotes'] == ''


def test_STR_matches_a_single_quoted_string_value():
    match = patterns.STR.match("'foo bar baz'   ")

    assert match is not None
    assert match.group() == "'foo bar baz'"

    groups = match.groupdict()
    assert groups['quotes'] == "'"


def test_STR_matches_a_double_quoted_string_value():
    match = patterns.STR.match('"foo bar baz" ')

    assert match is not None
    assert match.group() == '"foo bar baz"'

    groups = match.groupdict()
    assert groups['quotes'] == '"'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import codecs
import os

from setuptools import setup


def read(fname):
    file_path = os.path.join(os.path.dirname(__file__), fname)
    return codecs.open(file_path, encoding='utf-8').read()


setup(
    name='poyo',
    version='0.4.1',
    author='Raphael Pierzina',
    author_email='raphael@hackebrot.de',
    maintainer='Raphael Pierzina',
    maintainer_email='raphael@hackebrot.de',
    license='MIT',
    url='https://github.com/hackebrot/poyo',
    description='A lightweight YAML Parser for Python',
    long_description=read('README.rst'),
    packages=[
        'poyo',
    ],
    package_dir={'poyo': 'poyo'},
    include_package_data=True,
    zip_safe=False,
    install_requires=[],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Programming Language :: Python',
    ],
    keywords=['YAML', 'parser'],
)
